from sqlalchemy.orm import Query
from dozersim.suppliers import session
from dozersim.modelling import Settings


class DatabaseLink:
    """
    Adapter that couples settings (Settings) to a database table

    """

    def __init__(self, settings: Settings):
        self._settings: Settings = settings
        if settings.supplier_table:
            self._table = settings.supplier_table
        else:
            raise Exception("No supplier table linked to this setting class! Implement the supplier_table getter "
                            "method.")
        self._query: Query
        self.reset_query()

    def __iter__(self):
        return self._query.__iter__()

    def list_table(self):
        for entry in self._query:
            print(f"Table entry {entry.name}")

    def set_value(self, value: Query | str | Settings):
        while True:
            if type(value) is str:
                value = self._query.filter_by(name=value).first()
            else:
                for attr in [val for val in value.__dict__.keys() if '_sa_' not in val]:
                    setattr(self._settings, attr, getattr(value, attr))
                break

    @property
    def query(self) -> Query:
        return self._query

    @query.setter
    def query(self, query: Query):
        self._query = query

    def reset_query(self):
        self._query = session.query(self._table)




from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Iterable, TYPE_CHECKING, Any

import numpy as np

from dozersim.parameters import DatabaseLink
if TYPE_CHECKING:
    from dozersim.path import Path
    from dozersim.variables import Variables
    from dozersim.modelling import Model
    from dozersim.constraints import Constraint
    from dozersim.parameters import Parameter
    from dozersim.objectives import Objective


class Result:

    def __init__(self):
        self._variables: list[Variables] = []
        self._result_objects: list[ResultObject] = []
        self.id: int = 0

    def add_result_objects(self, load_case: int | None, variable_list: list[Variables] = None, result_objs: list[
        ResultObject] = None):
        # add load case label and store result objects and variables
        if variable_list is not None:
            for var in variable_list:
                var.load_case = load_case
                self._variables.append(var)
        if result_objs is not None:
            for res in result_objs:
                res.load_case = load_case
                self._result_objects.append(res)

    @property
    def constraints(self) -> list[ConstraintValue]:
        return [result_obj for result_obj in self._result_objects if type(result_obj) is ConstraintValue]

    @property
    def costs(self) -> list[ObjectiveValue]:
        return [result_obj for result_obj in self._result_objects if type(result_obj) is ObjectiveValue]

    def get_result_objects(self, load_case: str = None, path: Path = None) -> list[ResultObject]:
        res_list = self._result_objects
        if load_case is not None:
            res_list = list(filter(lambda obj: obj.load_case == load_case, res_list))
        if path is not None:
            res_list = list(filter(lambda obj: obj.path == path, res_list))
        return res_list

    def get_variables(self, load_case: str = None, path: Path = None) -> list[Variables]:
        var_list = self._variables
        if load_case is not None:
            var_list = list(filter(lambda obj: obj.load_case == load_case, var_list))
        if path is not None:
            var_list = list(filter(lambda obj: obj.parent == path, var_list))
        return var_list

    @property
    def load_cases(self) -> set[str]:
        return {obj.load_case for obj in self._result_objects}.union({obj.load_case for obj in self._variables})

    @property
    def paths(self) -> set[Path]:
        return {obj.path for obj in self._result_objects}.union({obj.parent for obj in self._variables})

    def activate(self):
        for obj in self._result_objects:
            if isinstance(obj, ParameterValue):
                obj.parent.default_vals = obj.value
                obj.parent.reset()

    @property
    def result_table(self, load_case: str = None, path: Path = None) -> (list[str], list[list[str]]):
        # Compile a flat list of all parameters, costs and constraints
        data = []
        headers = ['Type', 'Load case', 'Path', 'Name', 'Value', 'Limit']
        for result_obj in self.get_result_objects(load_case, path):
            value = f'{result_obj.value:.4f}' if isinstance(result_obj.value, float) else result_obj.value
            path_name = result_obj.path.name if result_obj.path is not None else 'None'
            data.append([type(result_obj).__name__, result_obj.load_case, path_name, result_obj.name, value, result_obj.limits])
        return headers, data


def get_non_iterable(parameters: list[Parameter]):
    return tuple(filter(lambda parameter: not isinstance(parameter, Iterable), parameters))


def get_iterable(parameters: list[Parameter]):
    return tuple(filter(lambda parameter: isinstance(parameter, Iterable), parameters))


class Analysis:

    def __init__(self):
        self.results: list[Result] = []
        self._active_id: int = 0

    def add_result(self, result: Result):
        self.results.append(result)

    @property
    def load_cases(self) -> list[str]:
        return list(self.results[-1].load_cases)

    def get_result_values(self, load_case: str,
                          target_obj: Parameter | Objective) -> list:
        values = []
        for result in self.results:
            for res in result.get_result_objects(load_case):
                if res.parent is target_obj:
                    values.append(res.value)
        return values

    def activate_set(self, idx: int):
        result = self.result_dict[idx]
        result.activate()
        self._active_id = idx

    @property
    def result_dict(self) -> dict[int, Result]:
        return {result.id: result for result in self.results}

    def get_fittest_set(self, load_case: str, objective: Objective) -> int:
        # idx = list(np.flatnonzero(cost == np.min(cost)))
        fittest_val = float('inf')
        fittest_set = None
        for idx, result in self.result_dict.items():
            for obj in result.get_result_objects(load_case):
                if obj.parent is objective and obj.evaluate() < fittest_val:
                    fittest_val = obj.evaluate()
                    fittest_set = idx
        return fittest_set

    @property
    def analysis_table(self) -> (list[str], list[list[str]]):
        data = []
        headers = ['Set ID']
        headers.extend([obj.name for obj in self.results[0].get_result_objects() if isinstance(obj, ParameterValue) or isinstance(obj, ObjectiveValue)])
        for result in self.results:
            row = [f'{result.id:003}']
            for obj in result.get_result_objects():
                if isinstance(obj, ParameterValue) or isinstance(obj, ObjectiveValue):
                    value = f'{obj.value:.4f}' if isinstance(obj.value, float) else obj.value
                    row.append(value)
            data.append(row)
        return headers, data




class ResultObject(ABC):

    def __init__(self, path: Path = None):
        self.path: Path | None = path
        self.load_case: str | None = None
        self.parent: Any = None

    @property
    @abstractmethod
    def value(self):
        pass

    @property
    @abstractmethod
    def limits(self):
        pass

    @property
    @abstractmethod
    def name(self):
        pass


class Load(ResultObject):

    def __init__(self, parent: Model, path: 'Path', force_axial: np.ndarray, force_radial: np.ndarray):
        super().__init__(path=path)
        self.force_axial = force_axial
        self.force_radial = force_radial
        self.parent = parent

    @property
    def value(self):
        return np.max([np.max(np.abs(self.force_radial)), np.max(np.max(self.force_axial))])

    @property
    def limits(self):
        return ''

    @property
    def name(self):
        return self.parent.name


class ConstraintValue(ResultObject):

    def __init__(self, path: Path, parent: Constraint, model: Model, value: float,
                 lb: float = float('-inf'), ub: float = float('inf')):
        super().__init__(path=path)
        self.parent = parent
        self.model = model
        self._value = value
        self.unit: str = ''
        self.lb = lb
        self.ub = ub

    @property
    def value(self):
        return self._value

    def evaluate(self):
        return np.max([self.value / self.ub - 1, self.value / self.lb - 1])

    @property
    def limits(self):
        return f'{self.lb:.4f} ... {self.ub:.4f}'

    @property
    def name(self):
        return f'{self.model.name}.{self.parent.name}'


class ObjectiveValue(ResultObject):

    def __init__(self, path: Path, parent: Objective, unit: str, value: float):
        super().__init__(path=path)
        self.parent = parent
        self._value = value
        self.unit: str = ''

    def evaluate(self):
        return self.value * self.parent.goal * self.parent.scale

    @property
    def value(self):
        return self._value

    @property
    def limits(self):
        return ''

    @property
    def name(self):
        return self.parent.name


class ParameterValue(ResultObject):

    def __init__(self, parent: Parameter, value: float, path: Path = None):
        super().__init__(path=path)
        self.parent = parent
        self._value = value
        self.unit: str = ''

    @property
    def name(self):
        return self.parent.name

    @property
    def limits(self):
        return self.parent.limits

    @property
    def value(self):
        if type(self.parent) is DatabaseLink:
            return self._value.name
        else:
            return self._value

    @value.setter
    def value(self, value: Any):
        self._value = value



from __future__ import annotations
from dataclasses import dataclass
from dozersim.path import Path
from dozersim.modelling import Model, Settings
from dozersim import constraints
from dozersim.components import tools


@dataclass(slots=True)
class TransformerSettings(Settings):
    name: str = None
    ratio: float = None
    max_effort: float = None
    efficiency: float = None
    mass: float = None


def evaluate_transformer(paths: tuple[Path], transformer: TransformerSettings, model: Model):
    path = paths[0]

    effort_output = path.variables.efforts
    flow_output = path.variables.flow

    model.constraints['peak_effort'].evaluate(model=model, path=path)

    path.transform(var_type=path.variables.domain, name=f"{model.name} input")

    speed_input = flow_output * transformer.ratio

    path.variables.add('flow', speed_input, model, "Transformer input")

    for effort in effort_output:
        path.variables.add('effort', effort.data / transformer.ratio, model, effort.name)

    friction_torque = tools.calculate_friction_torque(effort=path.variables.effort,
                                                      power=path.variables.power,
                                                      efficiency=transformer.efficiency
                                                      )

    path.variables.add('effort', friction_torque, model, "Efficiency loss")


class Transformer(Model):
    def __init__(self, name: str = 'transformer', settings=TransformerSettings(), eval_fun=evaluate_transformer):
        super().__init__(name=name, settings=settings, eval_fun=eval_fun)
        self.add_constraint(constraints.PathConstraint(name='peak_effort', settings=settings, ub='max_effort',
                                                       statistic='absmax', target='effort'))





from __future__ import annotations
from dataclasses import dataclass, field
import numpy as np
from dozersim.path import Path
from dozersim.modelling import Model, Settings


@dataclass(slots=True)
class TriangleLinkageSettings(Settings):
    name: str = None
    length_lever: float = None
    length_base: float = None
    initial_angle: float = None
    direction: int = None

    @property
    def supplier_table(self):
        return None


def evaluate_triangle_linkage(paths: list[Path], linkage: TriangleLinkageSettings, model: Model):
    """
        Evaluates a triangular linkage mechanism (a, b, c) assuming that the lever link (b) and settings link (c) are of
        fixed length and the angle between b and c (alpha) is main the output variable, thus link a is of variable
        length and is the (translational) input variable.
        Note the all models are inverse and therefore for from output to input

    Parameters
    ----------
    paths: Path that will be used for all operation on the variables
    linkage: Settings of the linkage mechanism
    model: used to leave a parent reference in variables

    Returns
    -------

    """
    path, = paths
    path.check('rotational')

    angle_out = path.variables.displacement
    torques_out = path.variables.efforts
    path.transform(var_type='translational', name="Linkage translation")

    length_c = linkage.length_base
    length_b = linkage.length_lever

    initial_length = np.sqrt(length_c ** 2 + length_b ** 2 - 2 * length_c * length_b * np.cos(linkage.initial_angle))
    angle_bc = linkage.initial_angle + angle_out*linkage.direction
    length_a = np.sqrt(length_c ** 2 + length_b ** 2 - 2 * length_c * length_b * np.cos(angle_bc))
    angle_ba = np.arccos((length_b ** 2 + length_a ** 2 - length_c ** 2) / (2 * length_a * length_b))
    translation_in = length_a - initial_length
    ratio = np.sin(angle_ba) * length_b

    path.variables.add('displacement', translation_in, model, "Linkage input")
    # paths.add_result_object(analysis.Load())
    for torque_out in torques_out:
        path.variables.add('effort', torque_out.data / ratio, model, torque_out.name)


class TriangleLinkage(Model):

    def __init__(self, name: str = 'linkage', settings=TriangleLinkageSettings(), eval_fun=evaluate_triangle_linkage):
        super().__init__(name=name, settings=settings, eval_fun=eval_fun)




from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from typing import Callable, Any, TYPE_CHECKING
from collections import deque

import itertools
# from trust_constr import minimize, NonlinearConstraint, LinearConstraint, Bounds
import numpy as np
from numpy.typing import ArrayLike
from scipy.optimize import NonlinearConstraint, minimize, Bounds

if TYPE_CHECKING:
    from dozersim.modelling import Model
    from dozersim.path import Path
    from dozersim.parameters import Parameter, SettingsParameter
    from dozersim.objectives import Objective

from dozersim.results import Result, Analysis, ParameterValue


def get_non_iterable(parameters: dict[Parameter, Any]):
    iterable_pars = []
    for parameter in parameters.keys():
        try:
            iter(parameter)
        except TypeError:
            iterable_pars.append(parameter)

    return tuple(iterable_pars)


def get_iterable(parameters: dict[Parameter, Any]):
    iterable_pars = []
    for parameter in parameters.keys():
        try:
            iter(parameter)
            iterable_pars.append(parameter)
        except TypeError:
            pass

    return tuple(iterable_pars)


@dataclass(slots=True)
class Interaction(ABC):
    name: str
    paths: list[Path]
    items: list[Model]


class System:
    def __init__(self):
        self.paths: set[Path] = set()
        self.interactions: dict[int, list[Interaction]] = {}

    def add_interaction(self, interactions: Interaction | list[Interaction], step: int = 0):
        interactions = [interactions] if type(interactions) is not list else interactions
        try:
            # if load step exists
            self.interactions[step].extend(interactions)
        except KeyError:
            # if load case not exists add new load case
            self.interactions.update({step: interactions})

        [[self.paths.add(path) for path in step.paths] for step in interactions]

    @property
    def load_cases(self):
        return list(self.interactions.keys())

    @property
    def interaction_table(self) -> (list[str], list[list[str]]):
        data = []
        header = ['Interaction', 'Item'] + [path.name for path in self.paths]
        for step in list(self.interactions.values())[-1]:
            item_links = ['x' if path in step.paths else '' for path in self.paths]
            for item in step.items:
                row = [step.name] + [item.name] + item_links
                data.append(row)
        return header, data


class Simulation:
    """ 
    Simulation object with all the basic function to solve a items tree for a set of load cases

        Methods
        -------------
        add_parameter: add a parameter to use in the optimization
        add_objective: add an objective function

    """

    def __init__(self, system: System):
        self._analysis: Analysis = Analysis()
        self._parameters: dict[Parameter, Any] = {}
        self._objectives: list[Objective] = []
        self._has_run: bool = False
        self.system = system

    def add_parameter(self, *parameters: Parameter):
        """ Add a parameter to the latest step """
        for parameter in parameters:
            self._parameters[parameter] = None

    def add_objective(self, *objectives: Objective):
        for objective in objectives:
            self._objectives.append(objective)

    def extract_analysis(self):
        """ Extract analysis from simulation.
        Note the analysis will be deleted!

        Returns
        ------------
        analysis (Analysis): analysis stored in this simulation.

        """
        analysis = self._analysis
        self._analysis = Analysis()
        self._has_run = False
        return analysis

    def clear_parameters(self):
        """
        Parameters will be removed.

        """
        self._parameters = dict()

    def clear_objectives(self):
        """
        Objectives will be removed.

        """
        self._objectives = list()

    def run(self):
        """
        This method executes the necessary system evaluations depending on the added parameters and objectives
         the analysis are stored in _analysis (a Analysis class).

        """
        if not self._has_run:
            # First, attach objective functions to system and sources
            for objective in self._objectives:
                objective.engage()
            # The evaluation order depends on the parameters that are added to the simulation
            if not self._parameters:
                # No parameters in simulation step, just one evaluation needed
                result = self.evaluate()
                self._analysis.add_result(result)
            else:
                # Parameters are present in simulation step
                if get_iterable(self._parameters):
                    # There are iterable parameters in this step
                    # get list of tuples that yield all combinations of parameters
                    combinations = list(itertools.product(*get_iterable(self._parameters)))
                else:
                    # There are no iterable parameters and thus no combinations
                    combinations = [()]
                # Now we start iterating the combinations
                n_comb = 1
                for combination in combinations:
                    # First evaluate system for iterable system
                    self.update_parameters(pars=get_iterable(self._parameters),
                                           values=combination)
                    if get_non_iterable(self._parameters):
                        # There are optimization parameters in this step
                        # create optimizer and optimize parameters
                        optimizer = Optimizer(eval_func=self.evaluate, update_fun=self.update_parameters,
                                              pars=get_non_iterable(self._parameters))
                        optimizer.optimize()
                    # Get result and store
                    result = self.evaluate()
                    result.id = n_comb
                    n_comb += 1
                    # store result list in collection
                    self._analysis.add_result(result)
            # After running the simulation all objective wil be detached
            for objective in self._objectives:
                objective.disengage()
            # Lock simulation until it has been reset
            self._has_run = True
        else:
            raise Exception("This simulation has already run and is locked until a reset() is given!")

    def update_parameters(self, pars: tuple[Parameter] = (), values: tuple = ()):
        for parameter, value in zip(pars, values):
            self._parameters[parameter] = value

    def evaluate(self) -> Result:
        """
            update the values for the parameters and evaluate system
        """
        # Set all paramter values
        for parameter, value in self._parameters.items():
            parameter.set_value(value)
        # create result object
        result = Result()
        # Solve the system for all load cases
        for load_case in self.system.load_cases:
            for interaction in self.system.interactions[load_case]:
                for item in interaction.items:
                    item.evaluate(paths=interaction.paths)
            for path in self.system.paths:
                variable_list, result_objects = path.unload()
                result.add_result_objects(load_case, variable_list, result_objects)
                path.reset()

            # create list of parameter value
            par_list = [ParameterValue(value=val, parent=par) for par, val in self._parameters.items()]
            result.add_result_objects(load_case=load_case, result_objs=par_list) if par_list else None
        return result


class Optimizer:
    """ 
    Optimizer optimizes a set of parameters for the objective(s)
    that are found in the items or load case trees
    
        Attributes
        ------------
        memory (bool): if true all optimization interactions are stored

        Methods
        ------------
        optimize: start parameter optimization

    """

    def __init__(self, eval_func: Callable, update_fun: Callable, pars: tuple[SettingsParameter]):
        self._parameters = pars
        self._eval_fun = eval_func
        self.update_fun = update_fun

    def _eval_constraints(self, par_values: ArrayLike):
        """evaluates the constraints of a system-load_case pair. It first updates the parameters, and then evaluates the constraints

        Parameters
        ----------
        par_values: ArrayLike
            Set of parameter values
        """
        self.update_fun(pars=self._parameters,
                        values=par_values)
        # self system
        results = self._eval_fun()

        constraint_values = [constraint.evaluate() for constraint in results.constraints]
        # print(f"Parameter values are {par_values} and constraint values are {constraint_values}")
        # get constraint status
        return constraint_values

    def _eval_cost(self, par_values: ArrayLike):
        """evaluates the cost of a system-load_case pair. It first updates the parameters, and then evaluates the cost

        Parameters
        ----------
        par_values: ArrayLike
            Set of parameter values
        """
        self.update_fun(pars=self._parameters,
                        values=par_values)
        # calculate the model
        results = self._eval_fun()

        # evaluate the costs
        result_scope = list(filter(lambda cost: cost.parent.optimize, results.costs))

        cost_values = [cost.evaluate() for cost in result_scope]

        # print(f"Parameter values are {par_values} and cost is {cost_values}")
        return sum(cost_values)

    def optimize(self):
        methods = ['trust-constr', 'SLSQP', 'COBYLA']

        lower_bound = []
        upper_bound = []
        init_guess = []
        for parameter in self._parameters:
            lower_bound.append(parameter.lb)
            upper_bound.append(parameter.ub)
            init_guess.append(parameter.x0)

        bounds = Bounds(lb=lower_bound, ub=upper_bound, keep_feasible=True)
        nonlinear_constraint = NonlinearConstraint(fun=self._eval_constraints, lb=-np.inf, ub=0, jac='3-point')
        x0 = np.array(init_guess)

        res = minimize(self._eval_cost, x0=x0, method=methods[1], constraints=nonlinear_constraint,
                       options={'verbose': 1, 'disp': False}, bounds=bounds)
        print(res)






from __future__ import annotations

import json
import os
from dataclasses import dataclass, field

import numpy as np

from dozersim.path import Path
from dozersim.modelling import Model, Settings

__location__ = os.path.realpath(
    os.path.join(os.getcwd(), os.path.dirname(__file__)))




@dataclass(slots=True)
class GaitDataSettings(Settings):
    Data: dict = None
    SubjectBodyMass: float = None
    Joint: str = None
    Speed: str = None

    def __post_init__(self):
        with open(os.path.join(__location__, '../data/gait_data.json')) as file_object:
            self.Data: dict = json.load(file_object)


def evaluate_gait_data(paths: list[Path], gaitdata: GaitDataSettings, model: Model):
    path, = paths

    path.check('rotational')
    time = gaitdata.Data["TimeNormal"]
    time_new = path.variables.time
    target = gaitdata.Joint + gaitdata.Speed
    torque_target = "Torque" + target
    angle_target = "Angle" + target
    speed_target = "Speed" + target
    torque = np.interp(x=time_new, xp=time, fp=gaitdata.Data[torque_target])
    angle = np.interp(x=time_new, xp=time, fp=gaitdata.Data[angle_target])
    speed = np.interp(x=time_new, xp=time, fp=gaitdata.Data[speed_target])
    acceleration = np.gradient(speed, time_new)
    '''
    data = {
            0: angle,
            1: speed,
            2: acceleration
    }
    '''
    # flow = Flow(name=_target, parent=model, data=data)
    # paths.variables.flows.append(flow)
    path.variables.add('velocity', values=speed, parent=model, name=torque_target)
    path.variables.add('torque', torque * gaitdata.SubjectBodyMass, model, torque_target)


class GaitData(Model):

    def __init__(self, name: str = 'gait data', settings=GaitDataSettings, eval_fun=evaluate_gait_data):
        super().__init__(name=name, settings=settings, eval_fun=eval_fun)



from __future__ import annotations

from typing import List, Any, TYPE_CHECKING
from abc import ABC, abstractmethod
from sqlalchemy.orm import Query
import numpy as np
import copy

from dozersim.suppliers import session

if TYPE_CHECKING:
    from dozersim.modelling import Settings


class Parameter(ABC):
    """
    Parameter abstract class

    """

    def __init__(self, name: str, default_vals: Any):
        self.name = name
        self.default_vals = default_vals

    def reset(self):
        self.set_value(self.default_vals)

    @abstractmethod
    def set_value(self, value: Any):
        pass

    @property
    @abstractmethod
    def limits(self):
        pass

    @property
    @abstractmethod
    def trace(self):
        pass


class SettingsParameter(Parameter):
    """ Parameter class stores a handle of a settings object.
    when x0 is set, the parameters becomes an optimization parameter
    when num is set, the simulation iterates the parameter

        Attributes
        -----------
        name (str): name of the parameter
        lb (float): lower bound
        ub (float): upper bound
        x0 (float): starting point
        num (int): number of interactions
    """

    def __init__(self, name: str, settings: List[Settings] | Settings, keywords: List[str] | str,
                 lb: float = -10e5, ub: float = 10e5, x0: float = None, num: int = None, scale: float = 1):
        """
            Parameter
        """
        self.lb = lb
        self.ub = ub
        self.x0 = x0
        self.num = num
        if x0 is not None and num is not None:
            raise Exception('x0 and num cannot be specified at the same time!')
        elif x0 is None and num is None:
            raise Exception('Either x0 or num have to be specified!')
        self.scale = scale
        if type(settings) is not list: settings = [settings]
        self._settings: List[Settings] = settings
        self._keywords: List[str] = []
        if type(keywords) is not list: keywords = [keywords]
        [self._keywords.append(keyword) for keyword in keywords]
        default_vals = read_from_settings(settings, keywords)
        super().__init__(name, default_vals)

    @property
    def trace(self):
        trace_set = set()
        trace_set.update([type(setting).__name__+'.'+keyword for setting, keyword in zip(self._settings,self._keywords)])
        return f"{'+'.join(list(trace_set))}"

    def __iter__(self):
        if self.num is not None:
            return np.linspace(self.lb, self.ub, self.num).__iter__()
        else:
            return None

    def set_value(self, value):
        write_to_settings(self._settings, self._keywords, value*self.scale)

    @property
    def limits(self):
        return f'{self.lb:.4f} ... {self.ub:.4f}'


def write_to_settings(settings: List[Settings], keywords: List[str], value: Any | List[Any]):
    if type(value) is list:
        for setting, keyword, value in zip(settings, keywords, value):
            setting.set_setting(keyword, value)
    else:
        for setting, keyword in zip(settings, keywords):
            setting.set_setting(keyword, value)


def read_from_settings(settings: List[Settings], keywords: List[str]) -> list:
    return [setting.get_setting(keyword) for setting, keyword in zip(settings, keywords)]


class DatabaseLink(Parameter):
    """ 
    Adapter that couples settings (Settings) to a database table
    
    """

    def __init__(self, settings: Settings, name: str = "New supplier link"):
        self._settings: Settings = settings
        if settings.supplier_table:
            self._table = settings.supplier_table
        else:
            raise Exception("No supplier table linked to this setting class! Implement the supplier_table getter "
                            "method.")
        self._query: Query
        self.reset_query()
        default_vals = copy.deepcopy(settings)
        super().__init__(name=name, default_vals=default_vals)

    def __iter__(self):
        return self._query.__iter__()

    @property
    def trace(self):
        return type(self._settings).__name__

    def list_table(self):
        for entry in self._query:
            print(f"Table entry {entry.name}")

    def set_value(self, value: Query | str | Settings):
        while True:
            if type(value) is str:
                value = self._query.filter_by(name=value).first()
            else:
                for attr in [val for val in value.__dict__.keys() if '_sa_' not in val]:
                    setattr(self._settings, attr, getattr(value, attr))
                break

    @property
    def query(self) -> Query:
        return self._query

    @query.setter
    def query(self, query: Query):
        self._query = query

    def reset_query(self):
        self._query = session.query(self._table)

    @property
    def limits(self):
        return ''




from __future__ import annotations
from dataclasses import dataclass
import numpy as np

from dozersim.modelling import Settings
from dozersim.modelling import Model
from dozersim.path import Path
from dozersim.suppliers import BatteryTable
from dozersim import constraints


@dataclass(slots=True)
class BatterySetting(Settings):
    name: str = None
    peak_current: float = None
    voltage: float = None
    capacity: float = None
    mass_spec: float = None
    series: int = None
    parallel: int = None
    SystemPath: Path = None
    desired_life: float = None

    @property
    def mass(self):
        return self.mass_spec

    @property
    def supplier_table(self):
        return BatteryTable


def evaluate_battery(paths: tuple[Path], battery: BatterySetting, model: Model):
    path = paths[0]

    time = path.variables.time
    current_battery = np.zeros(time.shape)
    voltage_battery = np.zeros(time.shape)

    system_path = battery.SystemPath

    if system_path in paths:
        for path in paths:
            if path is not system_path:
                path.check('electrical')
                voltage = path.variables.effort
                voltage_battery = np.maximum(np.array(voltage), voltage_battery)
                system_path.variables.add('current', path.variables.current, model, f"{path.name} Current")
    else:
        raise Exception('The system paths specified in the settings is not found in the paths attached to the items')

    system_path.variables.add('voltage', voltage_battery, model, f"System Voltage")
    if battery.peak_current:
        model.constraints['peak_current'].evaluate(model=model, path=system_path)

    power = system_path.variables.power
    power_pos = np.max([power, np.zeros(power.shape)], axis=0)
    required_energy = np.trapezoid(power, time) / time[-1]
    life = battery.capacity / required_energy

    model.constraints['battery_life'].evaluate(model, path, life)


class Battery(Model):
    def __init__(self, name: str = 'linkage', settings: BatterySetting = BatterySetting(), eval_fun: callable = evaluate_battery):
        super().__init__(name=name, settings=settings, eval_fun=eval_fun)
        self.add_constraint(constraints.PathConstraint(name='peak_current', settings=settings, ub='peak_current',
                                                       statistic='absmax', target='flow'))
        self.add_constraint(constraints.CustomConstraint(name='battery_life', settings=settings, lb='desired_life'))



from dozersim.components.spring import Spring
from dozersim.components.gearbox import Gearbox
from dozersim.components.electricmotor import ElectricMotor
from dozersim.components.servodrive import ServoDrive
from dozersim.components.ballbearing import BallBearing
from dozersim.modelling import Model


def create_actuator() -> list[Model]:
    return [Spring(),
            Gearbox(),
            ElectricMotor(),
            ServoDrive(),
            ]



from __future__ import annotations
from typing import TYPE_CHECKING
from abc import ABC, abstractmethod

from dozersim.utils import target_dict, statistic_dict

if TYPE_CHECKING:
    from dozersim.modelling import Model, Settings
    from dozersim.path import Path

from dozersim.results import ConstraintValue


class Constraint(ABC):

    def __init__(self, name: str, lb: str | float = None, ub: str | float = None, settings: Settings = None, scale: float = 1) -> None:
        self.lb = lb
        self.ub = ub
        self.name = name
        self._settings: Settings = settings
        self._optimize = True

    @abstractmethod
    def evaluate(self, model: Model, path: Path, value: float = None):
        pass

    @property
    def optimize(self):
        pass

    def get_bound_kws(self, model: Model) -> dict:
        bounds = dict()
        for kw in {'ub', 'lb'}:
            # check if lb and ub are defined
            if isinstance(self.__getattribute__(kw), str):
                try:
                    bound_val = model.settings.get_setting(self.__getattribute__(kw))
                    if bound_val is not None:
                        bounds[kw] = bound_val
                except AttributeError:
                    raise Exception('Settings object not present and/or keyword for bound incorrect!')
            elif self.ub is float:
                bounds[kw] = self.ub
        return bounds


class PathConstraint(Constraint):

    def __init__(self, name: str, statistic: str, target: str, settings: Settings,
                 lb: str | float = None, ub: str | float = None, scale: float = 1) -> None:
        super().__init__(name, lb, ub, settings, scale)
        self._target = target_dict[target]
        self._statistic = statistic_dict[statistic]
        self._optimize = True

    @property
    def optimize(self):
        return self._optimize

    @optimize.setter
    def optimize(self, do_optimize: bool):
        self._optimize = do_optimize

    def evaluate(self, model: Model, path: Path, value: float = None):

        values = self._target.evaluate(path)
        value = self._statistic.evaluate(values)
        kws = self.get_bound_kws(model)
        path.add_result_object(
            ConstraintValue(parent=self, model=model, value=value, path=path, **kws))


class CustomConstraint(Constraint):

    def __init__(self, name: str, settings: Settings = None, lb: str | float = None, ub: str | float = None,
                 scale: float = 1) -> None:
        super().__init__(name, lb, ub, settings, scale)
        self._settings: Settings = settings
        self._optimize = True

    @property
    def optimize(self):
        return self._optimize

    @optimize.setter
    def optimize(self, do_optimize: bool):
        self._optimize = do_optimize

    def evaluate(self, model: Model, path: Path, value: float = None):
        kws = self.get_bound_kws(model)
        path.add_result_object(
            ConstraintValue(parent=self, model=model, value=value, path=path, **kws))
